//////////////////////////////////////////////////////////////////
// Simple implementation of Adams-Bashforth 8
// Needs linking with math library (GCC option -lm)
// Copyright: (c) Petr Peringer, FIT TU Brno, 2018

#include <stdio.h>
#include <math.h>               // sin() - used only for error computation
#include <stdlib.h>             // strtod()

//////////////////////////////////////////////////////////////////
// System: circle test
#define SIZE 2                  // state vector size of the system
// "Dynamic section" = continuous system description
// Input:  t = time of step start,
//         y = state vector = integrator outputs
// Output: yin = vector of derivatives = integrators input values
void f(double t, double *y, double *yin)
{
    yin[0] = y[1];
    yin[1] = -y[0];
}

void RK4step(double t, double h, double *y);    // for start only

// Adams-Bashforth 8th order method step
// Parameters: t = start time,
//             h = step size,
//             y = state vector
// Output:  new state in y
// Warning: it does not update time
void AB8step(double t, double h, double *y)
{
    int i;
    static unsigned n = 0;
    static unsigned p = 0;
    static double fin[8][SIZE]; // f values

    // E: compute current f value
    p = (p + 1) % 8;
    f(t, y, fin[p]);            // yin = f(t, y(t))

    if (++n < 8) {
        // we need better accuracy for start
        const int NN=8;
        for(int i=0; i<NN; ++i) {
            RK4step(t, h/NN, y);       // START
            t+=h/NN;
        }
        return;
    }

    for (i = 0; i < SIZE; i++) {
        y[i] += h/120960 * (  434241 * fin[p][i]
                           - 1152169 * fin[(p - 1) % 8][i]
                           + 2183877 * fin[(p - 2) % 8][i]
                           - 2664477 * fin[(p - 3) % 8][i]
                           + 2102243 * fin[(p - 4) % 8][i]
                           - 1041723 * fin[(p - 5) % 8][i]
                           +  295767 * fin[(p - 6) % 8][i]
                           -   36799 * fin[(p - 7) % 8][i]);
    }
}


// experiment control
int main(int argc, char *argv[])
{
    double stepsize = 0.01;     // default step size
    if(argc>1)
        stepsize = strtod(argv[1],NULL);
    if(stepsize<1e-6||stepsize>0.1) {
        fprintf(stderr,"%s: step size out of bounds\n", argv[0]);
        return 1;
    }
    printf("# stepsize = %g\n", stepsize);
    double y[SIZE] = { 0, 1 };  // initial values
    double t = 0;
    while (t < 20) {
        printf("%9.6f % -11g % -11g  % e\n", t, y[0], y[1], y[0] - sin(t));
        AB8step(t, stepsize, y);        // make step
        t += stepsize;          // advance the simulation time
    }
    printf("%9.6f % -11g % -11g  % e\n", t, y[0], y[1], y[0] - sin(t));
    return 0;
}


//////////////////////////////////////////////////////////////////
// Runge-Kutta 4-th order method step for starting
// Parameters: t = start time,
//             h = step size,
//             y = state vector
// Output:  new state in y
// Warning: it does not update time
void RK4step(double t, double h, double *y)
{
    double yin[SIZE];           // integrator input = derivative
    double ys[SIZE];            // temporary state
    double kx[SIZE];            // accumulated 4 stages results

    int i;

    f(t, y, yin);               // yin = f(t, y(t))

    for (i = 0; i < SIZE; i++) {
        double k1 = h * yin[i]; // k1 = h * f(t, y(t))
        kx[i] = k1;
        ys[i] = y[i] + k1 / 2;
    }

    f(t + h / 2, ys, yin);      // yin = f(t+h/2, y(t)+k1/2)

    for (i = 0; i < SIZE; i++) {
        double k2 = h * yin[i]; // k2 = h * f(t+h/2, y(t)+k1/2)
        kx[i] += 2 * k2;
        ys[i] = y[i] + k2 / 2;
    }

    f(t + h / 2, ys, yin);      // yin = f(t+h/2, y(t)+k2/2)

    for (i = 0; i < SIZE; i++) {
        double k3 = h * yin[i]; // k3 = h * f(t+h/2, y(t)+k2/2)
        kx[i] += 2 * k3;
        ys[i] = y[i] + k3;
    }

    f(t + h, ys, yin);          // yin = f(t+h, y(t)+k3)

    for (i = 0; i < SIZE; i++) {
        double k4 = h * yin[i]; // k4 = h * f(t+h, y(t)+k3)
        kx[i] += k4;
        // Result:
        // y(t+h) = y(t) + k1/6 + k2/3 + k3/3 + k4/6;
        y[i] += kx[i] / 6;
    }
    // Return: y = new state at time t+h
}
