//////////////////////////////////////////////////////////////////
// Simple implementation of Adams-Bashforth-Moulton 12
// Warning: unstable!
// Needs linking with math library (GCC option -lm)
// Copyright: (c) Petr Peringer, FIT TU Brno, 2018

#include <stdio.h>
#include <math.h>               // sin() - used only for error computation
#include <stdlib.h>             // strtod()

//////////////////////////////////////////////////////////////////
// System: circle test
#define SIZE 2                  // state vector size of the system
// "Dynamic section" = continuous system description
// Input:  t = time of step start,
//         y = state vector = integrator outputs
// Output: yin = vector of derivatives = integrators input values
void f(double t, double *y, double *yin)
{
    yin[0] = y[1];
    yin[1] = -y[0];
}

void RK4step(double t, double h, double *y);    // for start only

#define ORDER 12

// Adams-Bashforth-Moulton 12th order method step
// Parameters: t = start time,
//             h = step size,
//             y = state vector
// Output:  new state in y
// Warning: it does not update time
void ABM12step(double t, double h, double *y)
{
    int i;
    static unsigned n = 0;
    static unsigned p = 0;
    static double fin[ORDER][SIZE]; // memory: f values

    // E: compute current f value
    p = (p + 1) % ORDER;
    f(t, y, fin[p]);            // yin = f(t, y(t))

    if (++n < ORDER) {
        // we need better accuracy for start
        const int NN=16;
        for(int i=0; i<NN; ++i) {
            RK4step(t, h/NN, y);       // START
            t+=h/NN;
        }
        return;
    }

    double ys[SIZE];
    // P: predictor
    for (i = 0; i < SIZE; i++) {
        ys[i] = y[i] + h/958003200.0 *
                                 (  4527766399.0 * fin[p][i]
                                -  19433810163.0 * fin[(p - 1) % ORDER][i]
                                +  61633227185.0 * fin[(p - 2) % ORDER][i]
                                - 135579356757.0 * fin[(p - 3) % ORDER][i]
                                + 214139355366.0 * fin[(p - 4) % ORDER][i]
                                - 247741639374.0 * fin[(p - 5) % ORDER][i]
                                + 211103573298.0 * fin[(p - 6) % ORDER][i]
                                - 131365867290.0 * fin[(p - 7) % ORDER][i]
                                +  58189107627.0 * fin[(p - 8) % ORDER][i]
                                -  17410248271.0 * fin[(p - 9) % ORDER][i]
                                +   3158642445.0 * fin[(p -10) % ORDER][i]
                                -    262747265.0 * fin[(p -11) % ORDER][i]);
    }

    // E: compute new value f_{n+1}
    p = (p + 1) % ORDER;
    f(t + h, ys, fin[p]);       // yin = f(t+h, y(t+h))

    // C: corector
    for (i = 0; i < SIZE; i++) {
        y[i] += h / 958003200.0 *
                         (   262747265.0 * fin[p][i]
                          + 1374799219.0 * fin[(p - 1) % ORDER][i]
                          - 2092490673.0 * fin[(p - 2) % ORDER][i]
                          + 3828828885.0 * fin[(p - 3) % ORDER][i]
                          - 5519460582.0 * fin[(p - 4) % ORDER][i]
                          + 6043521486.0 * fin[(p - 5) % ORDER][i]
                          - 4963166514.0 * fin[(p - 6) % ORDER][i]
                          + 3007739418.0 * fin[(p - 7) % ORDER][i]
                          - 1305971115.0 * fin[(p - 8) % ORDER][i]
                          +  384709327.0 * fin[(p - 9) % ORDER][i]
                          -   68928781.0 * fin[(p -10) % ORDER][i]
                          +    5675265.0 * fin[(p -11) % ORDER][i]);
    }
    p = (p - 1) % ORDER;
}


//////////////////////////////////////////////////////////////////
// Experiment control
int main(int argc, char *argv[])
{
    double stepsize = 0.01;     // default step size
    if(argc>1)
        stepsize = strtod(argv[1],NULL);
    if(stepsize<1e-6||stepsize>0.1) {
        fprintf(stderr,"%s: step size out of bounds\n", argv[0]);
        return 1;
    }
    printf("# stepsize = %g\n", stepsize);
    double y[SIZE] = { 0, 1 };  // initial values
    double t = 0;
    while (t < 20) {
        printf("%9.6f % -11g % -11g  % e\n", t, y[0], y[1], y[0] - sin(t));
        ABM12step(t, stepsize, y);        // make step
        t += stepsize;          // advance the simulation time
    }
    printf("%9.6f % -11g % -11g  % e\n", t, y[0], y[1], y[0] - sin(t));
    return 0;
}


//////////////////////////////////////////////////////////////////
// Runge-Kutta 4-th order method step for starting
// Parameters: t = start time,
//             h = step size,
//             y = state vector
// Output:  new state in y
// Warning: it does not update time
void RK4step(double t, double h, double *y)
{
    double yin[SIZE];           // integrator input = derivative
    double ys[SIZE];            // temporary state
    double kx[SIZE];            // accumulated 4 stages results

    int i;

    f(t, y, yin);               // yin = f(t, y(t))

    for (i = 0; i < SIZE; i++) {
        double k1 = h * yin[i]; // k1 = h * f(t, y(t))
        kx[i] = k1;
        ys[i] = y[i] + k1 / 2;
    }

    f(t + h / 2, ys, yin);      // yin = f(t+h/2, y(t)+k1/2)

    for (i = 0; i < SIZE; i++) {
        double k2 = h * yin[i]; // k2 = h * f(t+h/2, y(t)+k1/2)
        kx[i] += 2 * k2;
        ys[i] = y[i] + k2 / 2;
    }

    f(t + h / 2, ys, yin);      // yin = f(t+h/2, y(t)+k2/2)

    for (i = 0; i < SIZE; i++) {
        double k3 = h * yin[i]; // k3 = h * f(t+h/2, y(t)+k2/2)
        kx[i] += 2 * k3;
        ys[i] = y[i] + k3;
    }

    f(t + h, ys, yin);          // yin = f(t+h, y(t)+k3)

    for (i = 0; i < SIZE; i++) {
        double k4 = h * yin[i]; // k4 = h * f(t+h, y(t)+k3)
        kx[i] += k4;
        // Result:
        // y(t+h) = y(t) + k1/6 + k2/3 + k3/3 + k4/6;
        y[i] += kx[i] / 6;
    }
    // Return: y = new state at time t+h
}
