//////////////////////////////////////////////////////////////////
// Simple implementation of Adams-Bashforth-Moulton 5
// Needs linking with math library (GCC option -lm)
// Copyright: (c) Petr Peringer, FIT TU Brno, 2019

#include <stdio.h>
#include <math.h>               // sin() - used only for error computation

//////////////////////////////////////////////////////////////////
// System: circle test
#define SIZE 2                  // state vector size of the system
// "Dynamic section" = continuous system description
// Input:  t = time of step start,
//         y = state vector = integrator outputs
// Output: yin = vector of derivatives = integrators input values
void f(double t, double *y, double *yin)
{
    yin[0] = y[1];
    yin[1] = -y[0];
}

void RK4step(double t, double h, double *y);    // for start only

#define ORDER 5
//////////////////////////////////////////////////////////////////
// Adams-Bashforth-Moulton 5th order method step
// Parameters: t = start time,
//             h = step size,
//             y = state vector
// Output:  new state in y
// Warning: it does not update time
void ABM5step(double t, double h, double *y)
{
    int i;
    static unsigned n = 0;
    static unsigned p = 0;
    static double fin[ORDER][SIZE]; // memory: f values

    // E: compute current f value
    p = (p + 1) % ORDER;
    f(t, y, fin[p]);            // yin = f(t, y(t))

    if (++n < ORDER) {
        RK4step(t, h, y);       // START
        return;
    }

    double ys[SIZE];
    // P: predictor
    for (i = 0; i < SIZE; i++) {
        ys[i] = y[i] + h/720 * (  1901.0 * fin[p][i]
                                - 2774.0 * fin[(p - 1) % ORDER][i]
                                + 2616.0 * fin[(p - 2) % ORDER][i]
                                - 1274.0 * fin[(p - 3) % ORDER][i]
                                +  251.0 * fin[(p - 4) % ORDER][i]
                               );
    }

    // E: compute new value f_{n+1}
    p = (p + 1) % ORDER;
    f(t + h, ys, fin[p]);       // yin = f(t+h, y(t+h))

    // C: corector
    for (i = 0; i < SIZE; i++) {
        y[i] += h/720 * (  251.0 * fin[p][i]
                         + 646.0 * fin[(p - 1) % ORDER][i]
                         - 264.0 * fin[(p - 2) % ORDER][i]
                         + 106.0 * fin[(p - 3) % ORDER][i]
                         -  19.0 * fin[(p - 4) % ORDER][i]
                        );
    }
    p = (p - 1) % ORDER;
}


//////////////////////////////////////////////////////////////////
// Experiment control
int main()
{
    double y[SIZE] = { 0, 1 };
    double stepsize = 0.01;

    double t = 0;
    while (t < 20) {
        printf("%8.3f % -11g % -11g  % e\n", t, y[0], y[1], y[0] - sin(t));
        ABM5step(t, stepsize, y);       // make step
        t += stepsize;          // advance the simulation time
    }
    printf("%8.3f % -11g % -11g  % e\n", t, y[0], y[1], y[0] - sin(t));
    return 0;
}


//////////////////////////////////////////////////////////////////
// Runge-Kutta 4-th order method step for starting
// Parameters: t = start time,
//             h = step size,
//             y = state vector
// Output:  new state in y
// Warning: it does not update time
void RK4step(double t, double h, double *y)
{
    double yin[SIZE];           // integrator input = derivative
    double ys[SIZE];            // temporary state
    double kx[SIZE];            // accumulated 4 stages results

    int i;

    f(t, y, yin);               // yin = f(t, y(t))

    for (i = 0; i < SIZE; i++) {
        double k1 = h * yin[i]; // k1 = h * f(t, y(t))
        kx[i] = k1;
        ys[i] = y[i] + k1 / 2;
    }

    f(t + h / 2, ys, yin);      // yin = f(t+h/2, y(t)+k1/2)

    for (i = 0; i < SIZE; i++) {
        double k2 = h * yin[i]; // k2 = h * f(t+h/2, y(t)+k1/2)
        kx[i] += 2 * k2;
        ys[i] = y[i] + k2 / 2;
    }

    f(t + h / 2, ys, yin);      // yin = f(t+h/2, y(t)+k2/2)

    for (i = 0; i < SIZE; i++) {
        double k3 = h * yin[i]; // k3 = h * f(t+h/2, y(t)+k2/2)
        kx[i] += 2 * k3;
        ys[i] = y[i] + k3;
    }

    f(t + h, ys, yin);          // yin = f(t+h, y(t)+k3)

    for (i = 0; i < SIZE; i++) {
        double k4 = h * yin[i]; // k4 = h * f(t+h, y(t)+k3)
        kx[i] += k4;
        // Result:
        // y(t+h) = y(t) + k1/6 + k2/3 + k3/3 + k4/6;
        y[i] += kx[i] / 6;
    }
    // Return: y = new state at time t+h
}
