
// Simple (not optimized) implementation of Runge-Kutta 2 (midpoint)
// Needs linking with math library (GCC option -lm)
// Copyright: (c) Petr Peringer, FIT TU Brno, 2018

#include <stdio.h>
#include <math.h>               // sin() - used only for error computation

// System: circle test
#define SIZE 2                  // state vector size of the system
// "Dynamic section" = continuous system description
// Input:  t = time of step start,
//         y = state vector = integrator outputs
// Output: yin = vector of derivatives = integrators input values
void f(double t, double *y, double *yin)
{
    yin[0] = y[1];
    yin[1] = -y[0];
}

// Runge-Kutta 2nd order (midpoint) method step
// Parameters: t = start time,
//             h = step size,
//             y = state vector
// Output:  new state in y
// Warning: it does not update time
void RK2step(double t, double h, double *y)
{
    double yin[SIZE];           // integrator input = derivative
    double ystart[SIZE];        // initial state
    // 4 stages results:
    double k1[SIZE];
    double k2[SIZE];
    double k3[SIZE];

    int i;
    for (i = 0; i < SIZE; i++)  // save initial value
        ystart[i] = y[i];

    f(t, y, yin);               // yin = f(t, y(t))

    for (i = 0; i < SIZE; i++) {
        k1[i] = h * yin[i];     // k1 = h * f(t, y(t))
        y[i] = ystart[i] + k1[i] / 2;
    }

    f(t + h / 2, y, yin);       // yin = f(t+h/2, y(t)+k1/2)

    for (i = 0; i < SIZE; i++) {
        k2[i] = h * yin[i];     // k3 = h * f(t+h, y(t)+k1/2)
        // Result:
        // y(t+h) = y(t) + k2;
        y[i] = ystart[i] + k2[i];
    }
    // Return: y = new state at time t+h
}

// Experiment control
int main()
{
    double y[SIZE] = { 0, 1 };
    double stepsize = 0.01;

    double t = 0;
    while (t < 20) {
        printf("%8.3f % -11g % -11g  % e\n", t, y[0], y[1], y[0] - sin(t));
        RK2step(t, stepsize, y);        // make step
        t += stepsize;          // advance the simulation time
    }
    printf("%8.3f % -11g % -11g  % e\n", t, y[0], y[1], y[0] - sin(t));
    return 0;
}
